#include <array>
#include <gsl/gsl_rng.h>
#include <algorithm>
#include <iterator>
#include <math.h>
#include <stdio.h>
#include <iostream>
#include <string>
#include <cstdio>
#include "include/fileio.h"
#include "include/birthdeathprocesses.h"
using namespace std;

// main function for rescue simulation
bool rescue(unsigned int n, double u, unsigned long long *Nt0,
  double *la, double *mu, unsigned int seed) {

  // DECLARE+INITIALIZE VARIABLES
  // initialize variables for gillespie algorithm (see Barnes(2015) eq. 7.2)
  double t=0.0; // set time to zero
  double r1,r2;   // random numbers
  // auxiliary variables
  int r=0;  // next reaction index
  int i; // Individual Class with i mutated plasmids
  double sum_a_i; // aux. variable for gillespie algorithm
  double a0; // sum of a_i values
  // population size at the t=0
  unsigned int N0=0;
  for (size_t i = 0; i <= n; i++) N0+=Nt0[i];
  int N=N0; // pop. size
  unsigned long long Nvec[n+1];
  for (size_t i = 0; i <= n; i++) Nvec[i]=Nt0[i];
  // propensity of birth (i=0..n) and of death
  // (i=n+1,...,2*n+1) processes.
  double a[2*(n+1)]; // declare array for propensity values.
  // declare random distributions
  const gsl_rng_type * T;
  gsl_rng * rng1;
  T = gsl_rng_default;
  rng1 = gsl_rng_alloc (T);
  gsl_rng_set(rng1,seed);

  // SIMULATION

  // For the case that homozygous mutant cells are not resistant, return 0.
  if (mu[n]>=la[n]) {
    return 0;
  }

  while ((N!=0) && Nvec[n]<= log (.01) / log(mu[n]/la[n]) ) {
    // for (size_t i = 0; i <= n; i++) std::cout << Nvec[i] << ' '; std::cout << N << '\n'; // #XXX testing
    // refresh propensity values a[]
    for (int i = 0; i < 2*(n+1); i++) {
      /** birth **/ if (i<n+1)  { a[i]=Nvec[i]*la[i];
      // printf("B:%d:%d:%d:%.2f:%.2f " ,i,Nvec[i],i,la[i],a[i]); // #XXX testing
    }
      /** death **/ else { a[i]=Nvec[i-(n+1)]*mu[i-(n+1)];
      // printf("D:%d:%d:%d:%.2f:%.2f " ,i-(n+1),Nvec[i-(n+1)],i,mu[i-(n+1)],a[i]); // #XXX testing
    }
    }
    a0=0.0; for (auto& ai : a) a0 += ai; // sum of a[] of all processes
    // generate random numbers from uniform distribution
    r1=gsl_rng_uniform_pos(rng1);
    r2=gsl_rng_uniform_pos(rng1);
    // identify next reaction
    sum_a_i=0.0;r=0;
    do {sum_a_i += a[r];
      if (r2*a0<=sum_a_i) break;
      r++;}
      while(r<2*n+2);
    // printf(" r=%d  ",r); // #XXX testing
    // increase current time
    t+=1.0/a0*log(1.0/r1);
    // evaluate reaction r
    if (r<n+1) { // birth
      i=r;
      birth(Nvec, i, n, u, rng1);
      N++;
    }
    else { // death
      i=r-(n+1);
      death(Nvec, i, n);
      N--;
    }
    // delete one cell if N>N0
    if (N>N0) {
      a0=0.0; for (auto& Ni : Nvec) a0 += Ni;
      r2=gsl_rng_uniform_pos(rng1);
      sum_a_i=0.0;r=0;
      do {sum_a_i += Nvec[r];
        if (r2*a0<=sum_a_i) break;
        r++;}
        while(r<n+1);
      i=r;
      death(Nvec,i,n);
      N--;
    }

  // printf("\n" ); //XXX testing
  }
  // return : 0 <-> population extinction, 1 <-> population rescue
  return (N!=0);
}

int main(int argc, char const *argv[]) {

  rescueinput i=filetoinput(argv[1]);

  int NSim = stoi(argv[2]);
  double result=0.;
  for (size_t k = 0; k < NSim; k++) {
    bool b=rescue(i.n, i.u, i.Nt0, i.la, i.mu, k);
    result+= b;
  }
  result/=NSim;
  std::cout << result<<"\n";
  return 0;
}
